@@ -29,15 +29,15 @@ import json
 from multiprocessing import Process, Queue, Value
 from os import path
 try:
-  from queue import Empty
+  from queue import Empty, Full
 except ImportError:
-  from Queue import Empty
+  from Queue import Empty, Full
 import socket
 import traceback
 try:
-  from urllib.parse import urljoin, urlparse
+  from urllib.parse import urldefrag, urljoin, urlparse
 except ImportError:
-  from urlparse import urljoin, urlparse
+  from urlparse import urldefrag, urljoin, urlparse
 try:
   from urllib.request import urlopen
 except ImportError:
@@ -58,30 +58,27 @@ class Checker():
   # result in 403.
   HEADERS = {'User-Agent': '%s/%s' % (lnkckr.__name__, lnkckr.__version__)}
 
-  def __init__(self, src=None, jsonsrc=None):
+  def __init__(self):
 
     self.links = {}
     self.json_filename = None
-    self.load(src, jsonsrc)
 
-  def load(self, src=None, jsonsrc=None):
+  def load(self, src=None, jsonsrc=None, do_update=False, update_status=False):
     """Load links from a file
 
     src can be a URL (starting with http), file-like object, or a filename.
+    jsonsrc can be a file-like or a filename.
 
-    jsonsrc can be a file-like or a filename. jsonsrc is saved at
-    self.json_filename for save function if it's a filename and can be loaded
-    successfully.
+    "src.json" may be saved if src is a file and jsonsrc isn't a file.
 
-    jsonsrc processes first, it updates self.links and returns if something
-    comes from jsonsrc. If not then process with src, then if src + '.json'
-    exists, then load() will be called by itself with that filename.
+    do_update will update the JSON.
 
     The file format depends how self.process() is implemented."""
     if jsonsrc:
       self.json_filename = self.load_json(jsonsrc)
-      if self.json_filename:
+      if not do_update:
         return
+      json_links = self.links
 
     if not src:
       return
@@ -94,10 +91,11 @@ class Checker():
       elif hasattr(src, 'read'):
         f = src
       else:
-        self.json_filename = src + '.json'
-        if path.exists(self.json_filename):
-          self.load(None, src + '.json')
-          return
+        if not self.json_filename:
+          self.json_filename = src + '.json'
+          if not do_update and path.exists(self.json_filename):
+            self.load_json(self.json_filename)
+            return
         f = open(src, 'r')
 
       self.process(f)
@@ -105,6 +103,11 @@ class Checker():
       if f:
         f.close()
 
+    if jsonsrc and do_update:
+      file_links = self.links
+      self.links = json_links
+      self.update_links(file_links, update_status)
+
   def process(self, data):
     """Process the data from load()"""
     pass
@@ -151,14 +154,55 @@ class Checker():
     >>> c.links
     {'http://example.com': {'status': None, 'data': 'foobar'}}
     """
+    # skip blank fragment
+    if url == '#':
+      return
     d = {'status': None}
     d.update(data or {})
     self.links.update({url: d})
 
-  def check_url(self, url):
+  def update_links(self, links, update_status=False):
+    """Update self.links with links"""
+    urls = set(self.links.keys())
+    nurls = set(links.keys())
+
+    # remove removed links
+    for url in urls - nurls:
+      del self.links[url]
+    # add new links
+    for url in nurls - urls:
+      self.links[url] = links[url]
+    # update links
+    for url in urls & nurls:
+      self.update_link(url, links[url], update_status)
+
+  def update_link(self, url, new_link, update_status=False):
+    """Update self.links[url] with new_link
+
+    >>> c = Checker()
+    >>> url = 'http://example.com/'
+    >>> c.add_link(url)
+    >>> new_link = {'status': 123, 'foobar': 'blah'}
+    >>> c.update_link(url, new_link)
+    >>> c.links[url]
+    {'status': None, 'foobar': 'blah'}
+    >>> new_link['foobar'] = 'duh'
+    >>> c.update_link(url, new_link, update_status=True)
+    >>> c.links[url]
+    {'status': 123, 'foobar': 'duh'}
+    """
+    link = self.links[url]
+    status = link['status']
+    link.update(new_link)
+    if not update_status:
+      link['status'] = status
+
+  # =====
+
+  def check_url(self, url, frags=None):
     """Check a url
 
-    Returns (final status code in string, final redirection url)
+    Returns (final status code(s) in list of string, final redirect url)
 
     >>> c = Checker()
     >>> c.check_url('http://example.com') # doctest: +SKIP
@@ -169,6 +213,7 @@ class Checker():
     method = 'HEAD'
     start_url = url
     status = ''
+    statuses = []
 
     while not status:
       if url.startswith('//'):
@@ -188,18 +233,20 @@ class Checker():
         p = url_comp.path
         if url_comp.query:
           p += '?' + url_comp.query
-        if url_comp.fragment:
+        if frags:
           method = 'GET'
         conn.request(method, p, headers=self.HEADERS)
         resp = conn.getresponse()
-        if resp.status == 200 and url_comp.fragment:
+        if resp.status == 200 and frags:
           rbody = resp.read().decode('utf8')
-          if 'id="%s"' % url_comp.fragment in rbody:
-            status = '200'
-          elif "id='%s'" % url_comp.fragment in rbody:
-            status = '200'
-          else:
-            status = '###'
+          for frag in frags:
+            if 'id="%s"' % frag in rbody:
+              status = '200'
+            elif "id='%s'" % frag in rbody:
+              status = '200'
+            else:
+              status = '###'
+            statuses.append(status)
         elif 300 <= resp.status < 400:
           if redirs >= MAX_REDIRS:
             status = 'RRR'
@@ -220,6 +267,10 @@ class Checker():
 
     if start_url == url and not redirs:
       url = None
+    if frags:
+      if not statuses:
+        statuses = [status]*len(frags)
+      return statuses, url
     return status, url
 
   def check_worker(self, q, r, running):
@@ -227,14 +278,25 @@ class Checker():
     try:
       while not q.empty() or running.value:
         try:
-          url = q.get(timeout=0.01)
+          url, frags = q.get(timeout=0.01)
+          if frags:
+            statuses, rurl = self.check_url(url, frags)
+            for frag, status in zip(frags, statuses):
+              data = (url + '#' + frag, (status, rurl))
+              while data:
+                try:
+                  r.put(data, False, 0.01)
+                  data = None
+                except Full:
+                  pass
+            continue
           r.put((url, self.check_url(url)))
         except Empty:
           pass
     except KeyboardInterrupt:
       pass
 
-  def update_links(self, r):
+  def check_update_links(self, r):
 
     while not r.empty():
       url, data = r.get(block=False)
@@ -243,6 +305,32 @@ class Checker():
       link['status'] = status
       link['redirection'] = final_url
       self.do_update(url, link)
+      return True
+
+  def _check_groupby(self, urls):
+    """Generator to group by parts of url without the fragment
+
+    The fragments will collected as a tuple.
+
+    >>> c = Checker()
+    >>> urls = ['http://example.com']
+    >>> list(c._check_groupby(urls))
+    [('http://example.com', ())]
+    >>> urls = [
+    ...   'http://example.com',
+    ...   'http://example.com/foo#bar1',
+    ...   'http://example.com/foo#bar2',
+    ... ]
+    >>> list(c._check_groupby(urls)) # doctest: +NORMALIZE_WHITESPACE
+    [('http://example.com', ()),
+     ('http://example.com/foo', ('bar1', 'bar2'))]
+    >>> urls = ['#foo1', '#foo2']
+    >>> list(c._check_groupby(urls))
+    [('', ('foo1', 'foo2'))]
+    """
+    key = lambda item: item[0]
+    for url, g in groupby(sorted(map(urldefrag, urls), key=key), key):
+      yield url, tuple(filter(None, (item[1] for item in g)))
 
   def check(self, f=None):
     """Check links
@@ -252,7 +340,7 @@ class Checker():
     """
     default_timeout = socket.getdefaulttimeout()
     q = Queue(self.QUEUE_SIZE)
-    r = Queue(self.QUEUE_SIZE*2)
+    r = Queue(self.QUEUE_SIZE)
     running = Value('b', 1)
     workers = []
     for i in range(self.MAX_WORKERS):
@@ -262,22 +350,35 @@ class Checker():
       worker.start()
       workers.append(worker)
 
+    total = 0
+    count = 0
     try:
       if f is None:
         f = lambda item: item[1]['status'] is None
-      for idx, item in enumerate(filter(f, self.links.items()), start=1):
-        q.put(item[0])
-        self.update_links(r)
-        if idx % self.SAVE_INT == 0:
+      urls = (item[0] for item in filter(f, self.links.items()))
+      gurls = self._check_groupby(urls)
+      for idx, item in enumerate(gurls, start=1):
+        total += 1 if not item[1] else len(item[1])
+        while item:
+          try:
+            q.put(item, False, 0.01)
+            item = None
+          except Full:
+            count += 1 if self.check_update_links(r) else 0
+        count += 1 if self.check_update_links(r) else 0
+        if count % self.SAVE_INT == 0:
+          self.do_save()
+      while count < total:
+        count += 1 if self.check_update_links(r) else 0
+        if count % self.SAVE_INT == 0:
           self.do_save()
-      self.update_links(r)
     except KeyboardInterrupt:
       pass
 
     running.value = 0
     for worker in workers:
       worker.join()
-    self.update_links(r)
+    self.check_update_links(r)
     self.do_save()
     socket.setdefaulttimeout(default_timeout)
 
@@ -320,10 +421,15 @@ class Checker():
     print()
 
     links = self.links
-    for url, link in links.items():
-      self.print_report_link(url, link)
 
-    unchecked = sum(1 for link in links.values() if link['status'] is None)
+    unchecked = 0
+    key = lambda item: item[1]['status'] or '---'
+    for status, g in groupby(sorted(links.items(), key=key), key=key):
+      if status == '---':
+        unchecked = len(list(g))
+        continue
+      for url, link in sorted(g):
+        self.print_report_link(url, link)
     if unchecked:
       print('*** checking process is not finished, '
             '%d links have not been checked. ***' % unchecked)
